@@ -17,7 +17,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
  * USA
  */
 
@@ -68,19 +68,19 @@ get_handler_priority(gint *priority, PyObject *kwargs)
     }
     pos = 0;
     PyDict_Next(kwargs, &pos, &key, &val);
-    if (!PyString_Check(key)) {
+    if (!_PyUnicode_Check(key)) {
 	PyErr_SetString(PyExc_TypeError,
 			"keyword argument name is not a string");
 	return -1;
     }
 
-    if (strcmp(PyString_AsString(key), "priority") != 0) {
+    if (strcmp(_PyUnicode_AsString(key), "priority") != 0) {
 	PyErr_SetString(PyExc_TypeError,
 			"only 'priority' keyword argument accepted");
 	return -1;
     }
 
-    *priority = PyInt_AsLong(val);
+    *priority = _PyLong_AsLong(val);
     if (PyErr_Occurred()) {
 	PyErr_Clear();
 	PyErr_SetString(PyExc_ValueError, "could not get priority value");
@@ -125,7 +125,7 @@ pyglib_idle_add(PyObject *self, PyObject *args, PyObject *kwargs)
     handler_id = g_idle_add_full(priority,
 				 _pyglib_handler_marshal, data,
 				 _pyglib_destroy_notify);
-    return PyInt_FromLong(handler_id);
+    return _PyLong_FromLong(handler_id);
 }
 
 
@@ -165,7 +165,7 @@ pyglib_timeout_add(PyObject *self, PyObject *args, PyObject *kwargs)
     handler_id = g_timeout_add_full(priority, interval,
 				    _pyglib_handler_marshal, data,
 				    _pyglib_destroy_notify);
-    return PyInt_FromLong(handler_id);
+    return _PyLong_FromLong(handler_id);
 }
 
 static PyObject *
@@ -204,7 +204,7 @@ pyglib_timeout_add_seconds(PyObject *self, PyObject *args, PyObject *kwargs)
     handler_id = g_timeout_add_seconds_full(priority, interval,
                                             _pyglib_handler_marshal, data,
                                             _pyglib_destroy_notify);
-    return PyInt_FromLong(handler_id);
+    return _PyLong_FromLong(handler_id);
 }
 
 static gboolean
@@ -235,8 +235,9 @@ iowatch_marshal(GIOChannel *source,
 	res = FALSE;
     } else {
         if (ret == Py_None) {
-            if (PyErr_Warn(PyExc_Warning, "glib.io_add_watch callback returned None;"
-                           " should return True/False")) {
+            if (PyErr_Warn(PyExc_Warning,
+			   "glib.io_add_watch callback returned None; "
+                           "should return True/False")) {
                 PyErr_Print();
             }
         }
@@ -294,7 +295,7 @@ pyglib_io_add_watch(PyObject *self, PyObject *args, PyObject *kwargs)
 				     (GDestroyNotify)_pyglib_destroy_notify);
     g_io_channel_unref(iochannel);
     
-    return PyInt_FromLong(handler_id);
+    return _PyLong_FromLong(handler_id);
 }
 
 static PyObject *
@@ -374,7 +375,7 @@ pyglib_child_watch_add(PyObject *unused, PyObject *args, PyObject *kwargs)
         Py_INCREF(child_data->data);
     id = g_child_watch_add_full(priority, pid, child_watch_func,
                                 child_data, child_watch_dnotify);
-    return PyInt_FromLong(id);
+    return _PyLong_FromLong(id);
 }
 
 static PyObject *
@@ -391,7 +392,7 @@ pyglib_markup_escape_text(PyObject *unused, PyObject *args, PyObject *kwargs)
         return NULL;
 
     text_out = g_markup_escape_text(text_in, text_size);
-    retval = PyString_FromString(text_out);
+    retval = _PyUnicode_FromString(text_out);
     g_free(text_out);
     return retval;
 }
@@ -400,17 +401,75 @@ static PyObject *
 pyglib_get_current_time(PyObject *unused)
 {
     GTimeVal timeval;
-    double ret;
 
     g_get_current_time(&timeval);
-    ret = (double)timeval.tv_sec + (double)timeval.tv_usec * 0.000001;
-    return PyFloat_FromDouble(ret);
+    return pyglib_float_from_timeval(timeval);
+}
+
+static PyObject*
+pyglib_get_user_cache_dir(PyObject *self)
+{
+    const char *path = g_get_user_cache_dir();
+
+    if (path)
+        return _PyUnicode_FromString(path);
+    else {
+        Py_INCREF(Py_None);
+        return Py_None;
+    }
+}
+
+static PyObject*
+pyglib_get_user_config_dir(PyObject *self)
+{
+    const char *path = g_get_user_config_dir();
+
+    if (path)
+        return _PyUnicode_FromString(path);
+    else {
+        Py_INCREF(Py_None);
+        return Py_None;
+    }
+}
+
+static PyObject*
+pyglib_get_user_data_dir(PyObject *self)
+{
+    const char *path = g_get_user_data_dir();
+
+    if (path)
+        return _PyUnicode_FromString(path);
+    else {
+        Py_INCREF(Py_None);
+        return Py_None;
+    }
+}
+
+static PyObject *
+pyglib_get_user_special_dir(PyObject *unused, PyObject *args, PyObject *kwargs)
+{
+    static char *kwlist[] = { "directory", NULL };
+    guint directory;
+    const char *path;
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs,
+                                     "i:glib.get_user_special_dir", kwlist,
+                                     &directory))
+        return NULL;
+
+    path = g_get_user_special_dir(directory);
+    if (path)
+        return _PyUnicode_FromString(path);
+    else {
+        Py_INCREF(Py_None);
+        return Py_None;
+    }
 }
 
 static PyObject *
 pyglib_main_depth(PyObject *unused)
 {
-    return PyInt_FromLong(g_main_depth());
+    return _PyLong_FromLong(g_main_depth());
 }
 
 static PyObject *
@@ -424,7 +483,8 @@ pyglib_filename_display_name(PyObject *self, PyObject *args)
 	return NULL;
 
     display_name = g_filename_display_name(filename);
-    py_display_name = PyUnicode_DecodeUTF8(display_name, strlen(display_name), NULL);
+    py_display_name = PyUnicode_DecodeUTF8(display_name,
+					   strlen(display_name), NULL);
     g_free(display_name);
     return py_display_name;
 }
@@ -440,7 +500,8 @@ pyglib_filename_display_basename(PyObject *self, PyObject *args)
 	return NULL;
 
     display_basename = g_filename_display_basename(filename);
-    py_display_basename = PyUnicode_DecodeUTF8(display_basename, strlen(display_basename), NULL);
+    py_display_basename = PyUnicode_DecodeUTF8(display_basename,
+					       strlen(display_basename), NULL);
     g_free(display_basename);
     return py_display_basename;
 }
@@ -458,12 +519,13 @@ pyglib_filename_from_utf8(PyObject *self, PyObject *args)
 			  &utf8string, &utf8string_len))
 	return NULL;
 
-    filename = g_filename_from_utf8(utf8string, utf8string_len, NULL, &bytes_written, &error);
+    filename = g_filename_from_utf8(utf8string, utf8string_len,
+				    NULL, &bytes_written, &error);
     if (pyglib_error_check(&error)) {
         g_free(filename);
         return NULL;
     }
-    py_filename = PyString_FromStringAndSize(filename, bytes_written);
+    py_filename = _PyUnicode_FromStringAndSize(filename, bytes_written);
     g_free(filename);
     return py_filename;
 }
@@ -479,17 +541,19 @@ pyglib_get_application_name(PyObject *self)
         Py_INCREF(Py_None);
         return Py_None;
     }
-    return PyString_FromString(name);
+    return _PyUnicode_FromString(name);
 }
 
 static PyObject*
-pyglib_set_application_name(PyObject *self, PyObject *args)
+pyglib_set_application_name(PyObject *self, PyObject *arg)
 {
-    char *s;
-
-    if (!PyArg_ParseTuple(args, "s:glib.set_application_name", &s))
-        return NULL;
-    g_set_application_name(s);
+    if (!PyString_Check(arg)) {
+	PyErr_Format(PyExc_TypeError,
+		     "first argument must be a string, not '%s'",
+		     PyString_AS_STRING(PyObject_Repr(arg)));
+	return NULL;
+    }
+    g_set_application_name(PyString_AS_STRING(arg));
     Py_INCREF(Py_None);
     return Py_None;
 }
@@ -504,43 +568,84 @@ pyglib_get_prgname(PyObject *self)
         Py_INCREF(Py_None);
         return Py_None;
     }
-    return PyString_FromString(name);
+    return _PyUnicode_FromString(name);
 }
 
 static PyObject*
-pyglib_set_prgname(PyObject *self, PyObject *args)
+pyglib_set_prgname(PyObject *self, PyObject *arg)
 {
-    char *s;
-
-    if (!PyArg_ParseTuple(args, "s:glib.set_prgname", &s))
-        return NULL;
-    g_set_prgname(s);
+    if (!PyString_Check(arg)) {
+	PyErr_Format(PyExc_TypeError,
+		     "first argument must be a string, not '%s'",
+		     PyString_AS_STRING(PyObject_Repr(arg)));
+	return NULL;
+    }
+    g_set_prgname(PyString_AS_STRING(arg));
     Py_INCREF(Py_None);
     return Py_None;
 }
 
-
-static PyMethodDef pyglib_functions[] = {
-    { "spawn_async",
-      (PyCFunction)pyglib_spawn_async, METH_VARARGS|METH_KEYWORDS },
-    { "main_context_default",
-      (PyCFunction)pyglib_main_context_default, METH_NOARGS },
+static PyMethodDef _glib_functions[] = {
     { "idle_add",
-      (PyCFunction)pyglib_idle_add, METH_VARARGS|METH_KEYWORDS },
+      (PyCFunction)pyglib_idle_add, METH_VARARGS|METH_KEYWORDS,
+      "idle_add(callable, user_data=None, priority=None) -> source id\n"
+      "  callable receives (user_data)\n"
+      "Adds a callable to be called whenever there are no higher priority\n"
+      "events pending to the default main loop." },
     { "timeout_add",
-      (PyCFunction)pyglib_timeout_add, METH_VARARGS|METH_KEYWORDS },
+      (PyCFunction)pyglib_timeout_add, METH_VARARGS|METH_KEYWORDS,
+      "timeout_add(interval, callable, user_data=None,\n"
+      "            priority=None) -> source id\n"
+      "  callable receives (user_data)\n"
+      "Sets a callable be called repeatedly until it returns False." },
     { "timeout_add_seconds",
-      (PyCFunction)pyglib_timeout_add_seconds, METH_VARARGS|METH_KEYWORDS },
+      (PyCFunction)pyglib_timeout_add_seconds, METH_VARARGS|METH_KEYWORDS,
+      "timeout_add(interval, callable, user_data=None,\n"
+      "            priority=None) -> source_id\n"
+      "  callable receives (user_data)\n"
+      "Sets a callable be called repeatedly until it returns False.\n"
+      "Use this if you want to have a timer in the \"seconds\" range\n"
+      "and do not care about the exact time of the first call of the\n"
+      "timer, use this for more efficient system power usage." },
     { "io_add_watch",
-      (PyCFunction)pyglib_io_add_watch, METH_VARARGS|METH_KEYWORDS },
-    { "source_remove",
-      (PyCFunction)pyglib_source_remove, METH_VARARGS },
+      (PyCFunction)pyglib_io_add_watch, METH_VARARGS|METH_KEYWORDS,
+      "io_add_watch(fd, condition, callback, user_data=None) -> source id\n"
+      "  callable receives (fd, condition, user_data)\n"
+      "Arranges for the fd to be monitored by the main loop for the\n"
+      "specified condition. Condition is a combination of glib.IO_IN,\n"
+      "glib.IO_OUT, glib.IO_PRI, gio.IO_ERR and gio.IO_HUB.\n" },
     { "child_watch_add",
-      (PyCFunction)pyglib_child_watch_add, METH_VARARGS|METH_KEYWORDS },
-    { "markup_escape_text",
-      (PyCFunction)pyglib_markup_escape_text, METH_VARARGS|METH_KEYWORDS },
-    { "get_current_time",
-      (PyCFunction)pyglib_get_current_time, METH_NOARGS },
+      (PyCFunction)pyglib_child_watch_add, METH_VARARGS|METH_KEYWORDS,
+      "child_watch_add(pid, callable, user_data=None,\n"
+                       "priority=None) -> source id\n"
+      "  callable receives (pid, condition, user_data)\n"
+      "Sets the function specified by function to be called with the user\n"
+      "data specified by data when the child indicated by pid exits.\n"
+      "Condition is a combination of glib.IO_IN, glib.IO_OUT, glib.IO_PRI,\n"
+      "gio.IO_ERR and gio.IO_HUB." },
+    { "source_remove",
+      (PyCFunction)pyglib_source_remove, METH_VARARGS,
+      "source_remove(source_id) -> True if removed\n"
+      "Removes the event source specified by source id as returned by the\n"
+      "glib.idle_add(), glib.timeout_add() or glib.io_add_watch()\n"
+      "functions." },
+    { "spawn_async",
+      (PyCFunction)pyglib_spawn_async, METH_VARARGS|METH_KEYWORDS,
+      "spawn_async(argv, envp=None, working_directory=None,\n"
+      "            flags=0, child_setup=None, user_data=None,\n"
+      "            standard_input=None, standard_output=None,\n"
+      "            standard_error=None) -> (pid, stdin, stdout, stderr)\n"
+      "Execute a child program asynchronously within a glib.MainLoop()\n"
+      "See the reference manual for a complete reference." },
+    { "main_context_default",
+      (PyCFunction)pyglib_main_context_default, METH_NOARGS,
+      "main_context_default() -> a main context\n"
+      "Returns the default main context. This is the main context used\n"
+      "for main loop functions when a main loop is not explicitly specified." },
+    { "main_depth",
+      (PyCFunction)pyglib_main_depth, METH_NOARGS,
+      "main_depth() -> stack depth\n"
+      "Returns the depth of the stack of calls in the main context." },
     { "filename_display_name",
       (PyCFunction)pyglib_filename_display_name, METH_VARARGS },
     { "filename_display_basename",
@@ -550,13 +655,23 @@ static PyMethodDef pyglib_functions[] = {
     { "get_application_name",
       (PyCFunction)pyglib_get_application_name, METH_NOARGS },
     { "set_application_name",
-      (PyCFunction)pyglib_set_application_name, METH_VARARGS },
+      (PyCFunction)pyglib_set_application_name, METH_O },
     { "get_prgname",
       (PyCFunction)pyglib_get_prgname, METH_NOARGS },
     { "set_prgname",
-      (PyCFunction)pyglib_set_prgname, METH_VARARGS },
-    { "main_depth",
-      (PyCFunction)pyglib_main_depth, METH_NOARGS },
+      (PyCFunction)pyglib_set_prgname, METH_O },
+    { "get_current_time",
+      (PyCFunction)pyglib_get_current_time, METH_NOARGS },
+    { "get_user_cache_dir",
+      (PyCFunction)pyglib_get_user_cache_dir, METH_NOARGS },
+    { "get_user_config_dir",
+      (PyCFunction)pyglib_get_user_config_dir, METH_NOARGS },
+    { "get_user_data_dir",
+      (PyCFunction)pyglib_get_user_data_dir, METH_NOARGS },
+    { "get_user_special_dir",
+      (PyCFunction)pyglib_get_user_special_dir, METH_VARARGS },
+    { "markup_escape_text",
+      (PyCFunction)pyglib_markup_escape_text, METH_VARARGS|METH_KEYWORDS },
     { NULL, NULL, 0 }
 };
 
@@ -623,7 +738,6 @@ pyglib_register_version_tuples(PyObject *d)
 static void
 pyglib_register_constants(PyObject *m)
 {
-
     PyModule_AddIntConstant(m, "SPAWN_LEAVE_DESCRIPTORS_OPEN",
 			    G_SPAWN_LEAVE_DESCRIPTORS_OPEN);
     PyModule_AddIntConstant(m, "SPAWN_DO_NOT_REAP_CHILD",
@@ -703,6 +817,23 @@ pyglib_register_constants(PyObject *m)
 			    G_OPTION_ERROR_BAD_VALUE);
     PyModule_AddIntConstant(m, "OPTION_ERROR_FAILED",
 			    G_OPTION_ERROR_FAILED);
+ 
+    PyModule_AddIntConstant(m, "USER_DIRECTORY_DESKTOP",
+                            G_USER_DIRECTORY_DESKTOP);
+    PyModule_AddIntConstant(m, "USER_DIRECTORY_DOCUMENTS",
+                            G_USER_DIRECTORY_DOCUMENTS);
+    PyModule_AddIntConstant(m, "USER_DIRECTORY_DOWNLOAD",
+                            G_USER_DIRECTORY_DOWNLOAD);
+    PyModule_AddIntConstant(m, "USER_DIRECTORY_MUSIC",
+                            G_USER_DIRECTORY_MUSIC);
+    PyModule_AddIntConstant(m, "USER_DIRECTORY_PICTURES",
+                            G_USER_DIRECTORY_PICTURES);
+    PyModule_AddIntConstant(m, "USER_DIRECTORY_PUBLIC_SHARE",
+                            G_USER_DIRECTORY_PUBLIC_SHARE);
+    PyModule_AddIntConstant(m, "USER_DIRECTORY_TEMPLATES",
+                            G_USER_DIRECTORY_TEMPLATES);
+    PyModule_AddIntConstant(m, "USER_DIRECTORY_VIDEOS",
+                            G_USER_DIRECTORY_VIDEOS);
 
     PyModule_AddStringConstant(m, "OPTION_REMAINING",
 			       G_OPTION_REMAINING);
@@ -710,15 +841,11 @@ pyglib_register_constants(PyObject *m)
 			       (char*) g_quark_to_string(G_OPTION_ERROR));
 }
 
-DL_EXPORT(void)
-init_glib(void)
+PYGLIB_MODULE_START(_glib, "glib._glib")
 {
-    PyObject *m, *d;
+    PyObject *d = PyModule_GetDict(module);
 
-    m = Py_InitModule("glib._glib", pyglib_functions);
-    pyglib_register_constants(m);
-
-    d = PyModule_GetDict(m);
+    pyglib_register_constants(module);
     pyglib_register_api(d);
     pyglib_register_error(d);
     pyglib_register_version_tuples(d);
@@ -730,3 +857,4 @@ init_glib(void)
     pyglib_option_context_register_types(d);
     pyglib_option_group_register_types(d);
 }
+PYGLIB_MODULE_END
